// Sample PKI application developed using the dongle_template project.
// Performs raw RSA signature operation.

// Do not change these pragma settings during development
#pragma attribute("aid", "f0 00 00 ff")
#pragma attribute("name", "Myapp")
#pragma attribute("access_list", "0001")]

#include <multos.h>
#include <mi.h>
#include <ISO7816.h>
#include "data.h"

// Function to allow this app to delete itself on Trust Hub dongles
// enabled with the Developer Community issuer-id
extern void checkForDelete (BYTE *pPub);

// Function to handle halts caused by the MULTOS O/S for different security reasons.
void abendHandler (void)
{
	SW1 = 0x9F;
	// Abend reason codes defined in mi.h, MI_ABEND_*
	SW2 = miAbendReason();
	multosExit();
}

void
main (void)
{
	/* Define local variables first - note automatically initialised to zero */
	DWORD dwCheckSum;
	sInitPinObject pinObject;
	BYTE b;
	BYTE abBuffer[MODULUS_LENGTH];
	BOOL fRunningInP22;
	BYTE abP22BuildNumber[] = {0x00, 0x02, 0x00, 0x02, 0x00, 0x29};

	/* Whilst developing your app, always include this as the first function call in main */
	/* MUST BE REMOVED BEFORE DEPLOYMENT */
	checkForDelete(uPublic.abPublic);

	/* Note if the app is being run in a MULTOS ML5-P22 chip */
	if (multosGetConfigData(0x0A00,abBuffer))
	{
		fRunningInP22 = (multosCompareFixedLength(6,abP22BuildNumber,abBuffer) == MULTOS_BLOCK1_EQ_BLOCK2);
	}

	/* If running in P22, an abend handler can be specified */
	if (fRunningInP22)
	{
		miSetAbendFunction((WORD)abendHandler);
	}


	/* ========== Application starts here ========== */

	/* During development, enable in-chip debugging when CLA = 0x81 */
	/* MUST BE REMOVED BEFORE DEPLOYMENT */
	if (fRunningInP22 && CLA == 0x81)
		miDebuggingEnable();

	/* Check CLA is OK */
	else if (CLA != 0x80)
		multosExitSW(ISO7816_SW_CLA_NOT_SUPPORTED);

	/* If PIN checking is activated */
	if (fStPinCheckingEnabled)
	{
		/* If the PIN hasn't been initialised */
		if (multosGetPINStatus() == PIN_NOT_INITIALISED)
		{
			/* If this command is not Initialise PIN */
			if (INS != INS_INIT_PIN)
			{
				/* Exit with status to indicate that the command isn't allowed yet */
				multosExitSW(ISO7816_SW_COMMAND_NOT_ALLOWED);
			}
		}

		/* If the command is NOT Verify PIN */
		else if (INS != INS_VERIFY_PIN)
		{
			/* Is the PIN is verified? */
			if (multosGetPINVerificationStatus() != PIN_VERIFIED)
			{
				/* No - exit with status to indicate that permission has not been granted */
				multosExitSW(ISO7816_SW_SECURITY_STATUS_NOT_SATISFIED);
			}
		}
	}

	/* If the RSA key has been generated */
	if (sStKeyStorage.fGenerated)
	{
		/* Calculate the checksum of the key storage */
		dwCheckSum = multosChecksum((BYTE*)&sStKeyStorage,sizeof(sKeyStorage));

		/* If the key storage checksum has changed */
		if (dwCheckSum != dwStKeyStorageChecksum)
		{
			/* Exit with status to indicate the key data has been corrupted */
			multosExitSW(ISO7816_SW_FILE_INVALID);
		}
	}

	/* Now handle each instruction */
	switch (INS)
	{
		case INS_INIT_PIN:

			/* Check command data */
			if (!multosCheckCase(CASE_DATA_IN))
				multosExitSW(ISO7816_SW_WRONG_DATA);
			if (Lc != sizeof(sInitPinData))
				multosExitSW(ISO7816_SW_WRONG_LENGTH);

			/* Check the PIN is of the correct length */
			if (uPublic.pinData.bPinLen != PIN_LENGTH)
			{
				/* Exit indicating that the parameter supplied is incorrect */
				multosExitSW(ISO7816_SW_WRONG_PARAMS);
			}

			/* Set up PIN initialisation object */
			multosCopy((BYTE*)&(uPublic.pinData),(BYTE*)&(pinObject.data),sizeof(sInitPinData));
			pinObject.dwChecksum = multosChecksum((BYTE*)&(pinObject.data),sizeof(pinObject.data));

			/* Initialise the PIN */
			multosInitialisePIN((BYTE*)&pinObject);

			/* Check if the initialisation was successful */
			if (multosGetPINStatus() != PIN_INITIALISED)
			{
				/* Exit indicating that an error has occurred */
				multosExitSW(ISO7816_SW_UNKNOWN);
			}

			// The PIN must be verified from now on.
			fStPinCheckingEnabled = TRUE;
			break;

		case INS_VERIFY_PIN:

			/* Check command data */
			if (!multosCheckCase(CASE_DATA_IN))
				multosExitSW(ISO7816_SW_WRONG_DATA);
			if (Lc != PIN_LENGTH)
				multosExitSW(ISO7816_SW_WRONG_LENGTH);

			/* Is the PIN blocked? */
			b = multosGetPINTryCounter();
			if (b == 0)
			{
				/* Exit indicating that the app is now blocked */
				multosExitSW(ISO7816_SW_FUNC_NOT_SUPPORTED);
			}

			/* Is PIN checking required yet? */
			if (!fStPinCheckingEnabled)
			{
				/* Exit indicating that the PIN isn't set up yet */
				multosExitSW(ISO7816_SW_FILE_NOT_FOUND);
			}

			/* Decrement the PIN try counter - it will be restored if verification is successful */
			b--;
			multosSetPINTryCounter(b);

			/* Attempt to verify the PIN */
			if (multosVerifyPIN(uPublic.abPin,PIN_LENGTH) != PIN_VERIFICATION_OK)
			{
				/* Change the PIN status to not verified */
				multosSetPINVerificationStatus(PIN_UNVERIFIED);

				/* Exit with status to indicate that permission has not been granted */
				multosExitSW(ISO7816_SW_SECURITY_STATUS_NOT_SATISFIED);
			}

			/* Change the PIN status to verified */
			multosSetPINVerificationStatus(PIN_VERIFIED);

			/* Restore the PIN Try Counter */
			multosSetPINTryCounter( multosGetPINTryLimit() );

			break;

		case INS_GEN_KEYPAIR:

			/* Check command data */
			if (!multosCheckCase(CASE_DATA_OUT))
				multosExitSW(ISO7816_SW_WRONG_DATA);

			/* If the key has already been generated */
			if (sStKeyStorage.fGenerated)
			{
				/* Mark as no longer generated */
				sStKeyStorage.fGenerated = FALSE;
			}

			/* Attempt to generate the key pair */
			sStKeyStorage.keyPair.abExponent[0] = 3;
			b = multosGenerateRsaKeyPair(
						KEYGEN_METHOD_DEFAULT,
						KEYGEN_MODE_BALANCED,
						MODULUS_LENGTH,
						sStKeyStorage.keyPair.abPublicKey,
						sStKeyStorage.keyPair.abPrivateKey,
						sStKeyStorage.keyPair.abExponent,
						EXPONENT_LENGTH,
						MODULUS_LENGTH);

			/* If the key generated OK */
			if (b)
			{
				/* Mark as being generated */
				sStKeyStorage.fGenerated = TRUE;

				/* Update the key store checksum */
				dwStKeyStorageChecksum = multosChecksum((BYTE*)&sStKeyStorage,sizeof(sKeyStorage));

				/* Output the public key */
				multosCopy(sStKeyStorage.keyPair.abPublicKey,uPublic.abPublic,MODULUS_LENGTH);
				La = MODULUS_LENGTH;
			}
			else
			{
				/* Return a status to indicate that the operation failed */
				multosExitSW(ISO7816_SW_UNKNOWN);
			}

			// The PIN must be verified from now on.
			fStPinCheckingEnabled = TRUE;

			break;

		case INS_SIGNATURE:
			/* Check command data */
			if (!multosCheckCase(CASE_DATA_INOUT))
				multosExitSW(ISO7816_SW_WRONG_DATA);
			if (Lc != MODULUS_LENGTH)
				multosExitSW(ISO7816_SW_WRONG_LENGTH);
			if (Le > MODULUS_LENGTH)
				multosExitSW(ISO7816_SW_WRONG_LENGTH);

			/* If the keypair hasn't been generated */
			if (!sStKeyStorage.fGenerated)
			{
				/* Exit with an error indicating that the application is not ready to use */
				multosExitSW(ISO7816_SW_CONDITIONS_NOT_SATISFIED);
			}

			/* Do the signature */
			multosModularExponentiationCRT(
						abBuffer,		// Output
						uPublic.abPublic,	// Input
						sStKeyStorage.keyPair.abPrivateKey+MODULUS_LENGTH,	// Pointer to dp | dq
						sStKeyStorage.keyPair.abPrivateKey,					// Pointer to p | q | u
						MODULUS_LENGTH);

			/* Calculate how much of the signature to return */
			if (Le == 0)
				La = MODULUS_LENGTH;
			else
				La = Le;

			/* Copy that much of the signature to the output */
			multosCopy(abBuffer,uPublic.abPublic,La);
			break;

		default:
			multosExitSW(ISO7816_SW_INS_NOT_SUPPORTED);

	}
}


